# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
from minio import ResponseError

from matorage.connector import MTRConnector


class Uploader(MTRConnector):
    r""" File Storage uploader class with multi thread.
        MinIO is thread-safety, according to document.
        Although Python Global Interpreter Lock(GIL), multi thread can benefit greatly from file IO.
    """

    def __init__(
        self, client, bucket, num_worker_threads, multipart_upload_size, inmemory=False
    ):
        super(Uploader, self).__init__(client, bucket, num_worker_threads)
        self._multipart_upload_size = multipart_upload_size
        self._inmemory = inmemory

    def do_job(self, local_file, remote_file):

        if isinstance(remote_file, str):
            minio_key = remote_file
        try:
            if not self._inmemory:
                self._client.fput_object(
                    self._bucket,
                    minio_key,
                    local_file,
                    part_size=self._multipart_upload_size,
                )
                os.remove(local_file)
            else:
                fileimage = io.BytesIO(local_file)
                self._client.put_object(
                    self._bucket,
                    minio_key,
                    fileimage,
                    len(local_file),
                    part_size=self._multipart_upload_size,
                )
        except ResponseError as err:
            print(err)



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from minio import ResponseError

from matorage.connector import MTRConnector


class Downloader(MTRConnector):
    r""" File Storage downloader class with multi thread.
        MinIO is thread-safety, according to document.
        Although Python Global Interpreter Lock(GIL), multi thread can benefit greatly from file IO.
    """

    def __init__(self, client, bucket, num_worker_threads):
        super(Downloader, self).__init__(client, bucket, num_worker_threads)

    def do_job(self, local_file, remote_file):
        if isinstance(remote_file, str):
            minio_key = os.path.basename(remote_file)
        try:
            self._client.fget_object(self._bucket, minio_key, local_file)
        except ResponseError as err:
            print(err)



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest
from minio import Minio
from urllib.parse import urlsplit

from matorage.nas import NAS


class OptimizerTest(unittest.TestCase):
    optimizer_config = None
    optimizer_config_file = None
    optimizer_manager = None
    storage_config = {
        "endpoint": "127.0.0.1:9001",
        "access_key": "minio",
        "secret_key": "miniosecretkey",
        "secure": False,
    }
    nas_config = {
        "endpoint": "/tmp/unittest",
    }

    def check_nas(self, endpoint):
        _url_or_path = "//" + endpoint
        u = urlsplit(_url_or_path)
        if u.path:
            return True
        if u.netloc:
            return False
        raise ValueError("This endpoint is not suitable.")

    def tearDown(self):
        if self.optimizer_manager is not None:
            # delete bucket
            client = (
                Minio(**self.storage_config)
                if not self.check_nas(self.optimizer_config.endpoint)
                else NAS(self.optimizer_config.endpoint)
            )
            objects = client.list_objects(
                self.optimizer_config.bucket_name, recursive=True
            )
            for obj in objects:
                client.remove_object(self.optimizer_config.bucket_name, obj.object_name)
            client.remove_bucket(self.optimizer_config.bucket_name)

        if self.optimizer_config_file is not None:
            if os.path.exists(self.optimizer_config_file):
                os.remove(self.optimizer_config_file)



# Copyright 2020-present Tae Hwan Jung
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup
from sutils import find_name, get_setuptools, check_torch_tf_version

project_name = "matorage"
version = os.environ.get('MATORAGE_VERSION', '0.0.0')

if __name__ == "__main__":
    check_torch_tf_version()

    project_name = find_name()

    with open('README.md', 'r') as t:
        README = t.read()

    setup(
        # Project Name, Version
        name=project_name,
        version=version,
        long_description=README,
        long_description_content_type='text/markdown',
        # Author
        license="Apache License, Version 2.0",
        author="TaeHwan-Jung",
        author_email="nlkey2022@gmail.com",
        description="matorage is Matrix or Tensor(multidimensional matrix) "
                    "Object Storage with high availability "
                    "distributed systems for Deep Learning framework.",
        url="https://github.com/graykode/matorage",
        # Platform, Requires
        python_requires=">=3.5",
        platforms=["any"],
        project_urls={
            "Documentation": "https://matorage.readthedocs.io/en/stable/",
            "Source Code": "https://github.com/graykode/matorage",
        },
        **get_setuptools()
    )


# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import tables
import hashlib
import tempfile
from minio import Minio

from matorage.nas import NAS
from matorage.utils import check_nas, logger
from matorage.uploader import Uploader

_KB = 1024
"""The size of a Kilobyte in bytes"""

_MB = 1024 * _KB
"""The size of a Megabyte in bytes"""


class Manager(object):
    type = "model"

    def __init__(self, config, num_worker_threads=4, multipart_upload_size=5 * _MB):
        self.config = config
        self.num_worker_threads = num_worker_threads
        self.multipart_upload_size = multipart_upload_size

        self._client = (
            Minio(
                endpoint=self.config.endpoint,
                access_key=self.config.access_key,
                secret_key=self.config.secret_key,
                secure=self.config.secure,
                region=self.config.region,
            )
            if not check_nas(self.config.endpoint)
            else NAS(self.config.endpoint)
        )

        self._uploader = Uploader(
            client=self._client,
            bucket=self.config.bucket_name,
            num_worker_threads=self.num_worker_threads,
            multipart_upload_size=self.multipart_upload_size,
            inmemory=True,
        )

    def _uploader_closing(self):
        self._uploader.join_queue()

        _metadata_file = tempfile.mktemp("metadata.json")
        with open(_metadata_file, "w", encoding="utf-8") as writer:
            writer.write(
                json.dumps(self.config.metadata, indent=4, sort_keys=True) + "\n"
            )

        self._client.fput_object(
            bucket_name=self.config.bucket_name,
            object_name="metadata.json",
            file_path=_metadata_file,
        )
        os.remove(_metadata_file)

    def _save_with_clear(self, model_folder, model, overwrite=False):
        if overwrite:
            objects = self._client.list_objects(
                bucket_name=self.config.bucket_name, prefix=f"{model_folder}/"
            )
            for obj in objects:
                self._client.remove_object(
                    bucket_name=self.config.bucket_name, object_name=obj.object_name
                )

        # saving model
        self._save_model(model_folder, model)
        self._uploader_closing()

    def _save_layer(self, model_folder, name, weight):
        _local_file = tempfile.mktemp(f"{name}.h5")

        _file = tables.open_file(
            _local_file, "w", driver="H5FD_CORE", driver_core_backing_store=False
        )
        _file.create_carray(
            "/", self.type, obj=weight, filters=tables.Filters(**self.config.compressor)
        )

        self._uploader.set_queue(
            local_file=_file.get_file_image(), remote_file=f"{model_folder}/{name}"
        )
        _file.close()

    def save(self, model, **kwargs):
        if not self._client.bucket_exists(self.config.bucket_name):
            self._client.make_bucket(
                self.config.bucket_name, location=self.config.region
            )

        if not isinstance(kwargs, dict):
            metadata = 0
        else:
            metadata = kwargs

        model_folder = self._hashmap_transfer(metadata)

        if model_folder in self.config.metadata["model"]:
            logger.info(
                "{} {} is already exist, so model will be overwrited.".format(
                    self.config.model_name, str(self.config.additional)
                )
            )
            self._save_with_clear(model_folder, model, overwrite=True)
        else:
            self.config.metadata["model"].update({model_folder: metadata})
            self._save_with_clear(model_folder, model)

        logger.info("model with {} is saved".format(str(metadata)))

    def load(self, model, **kwargs):
        if not isinstance(kwargs, dict):
            metadata = 0
        else:
            metadata = kwargs

        model_folder = self._hashmap_transfer(metadata)

        layers = self._client.list_objects(
            bucket_name=self.config.bucket_name,
            prefix=f"{model_folder}/",
            recursive=True,
        )

        logger.info("model with {} is loaded".format(str(metadata)))
        return self._load_model(model_folder, layers, model)

    def _hashmap_transfer(self, metadata):
        """
        Get unikey object folder with `metadata` of model.

        Returns:
            :obj: `str`:
        """
        if isinstance(metadata, int):
            metadata = str(metadata)
        if not isinstance(metadata, str) and not isinstance(metadata, dict):
            raise ValueError(
                "metadata {} is empty or not str and dict type".format(metadata)
            )

        key = json.dumps(metadata, indent=4, sort_keys=True)
        return hashlib.md5(key.encode("utf-8")).hexdigest()

    @property
    def get_metadata(self):
        """
        Get all models according to metadata(ex. step, epoch)

        Returns:
            :obj:`dict`: model of metadata

        Examples::

            >>> model_manager.save(model, step=100)
            >>> model_manager.save(model, step=200)
            >>> model_manager.get_metadata
            {
                'additional': {'version': '1.0.1'},
                'compressor': {'complevel': 0, 'complib': 'zlib'},
                'endpoint': '127.0.0.1:9000',
                'model':
                {
                    'ad44168f1343bc77b4d9ad6f1fef50b6': {'step': 100},
                    'af0677ecf0d15d17d10204be9ff2f9f5': {'step': 200}
                },
                'model_name': 'testmodel'
            }

        """
        return self.config.metadata



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_KB = 1024
"""The size of a Kilobyte in bytes"""

_MB = 1024 * _KB
"""The size of a Megabyte in bytes"""

import io
import ast
import h5py
import json
import numpy as np
import tensorflow as tf
from tensorflow.python.keras.optimizer_v2.optimizer_v2 import OptimizerV2

from matorage.optimizer.manager import Manager


class OptimizerManager(Manager):

    """
        Optimizer Manager Tensorflow classes. This class overrides ``Manager``.

        Note:
            Unlike Dataset, optimizer weight is loaded entirely into cpu memory.
            Therefore, the `HDF5_CORE` driver using the memory option is default setting.

        Args:
            config (:obj:`matorage.OptimizerConfig`, **require**):
                A OptimizerConfig instance object
            num_worker_threads (:obj:`int`, optional, defaults to `4`):
                Number of backend storage worker to upload or download.
            multipart_upload_size (:obj:`integer`, optional, defaults to `5 * 1024 * 1024`):
                size of the incompletely uploaded object.
                You can sync files faster with `multipart upload in MinIO. <https://github.com/minio/minio-py/blob/master/minio/api.py#L1795>`_
                This is because MinIO clients use multi-threading, which improves IO speed more
                efficiently regardless of Python's Global Interpreter Lock(GIL).

        Examples::

            from matorage.tensorflow import OptimizerManager
            optimizer_config = OptimizerConfig(
                endpoint='127.0.0.1:9000',
                access_key='minio',
                secret_key='miniosecretkey',
                optimizer_name='testoptimizer',
                additional={
                    "version" : "1.0.1"
                }
            )

            optimizer_manager = OptimizerManager(config=optimizer_config)
            optimizer_manager.save(model.optimizer, step=100)

        """

    def __init__(self, config, num_worker_threads=4, multipart_upload_size=5 * _MB):
        super(OptimizerManager, self).__init__(
            config, num_worker_threads, multipart_upload_size
        )

    def _get_step(self, optimizer):
        if len(optimizer.weights) > 0:
            return optimizer.weights[0].numpy()
        else:
            return None

    def _set_metadata(self, metadata, optimizer, step):
        assert isinstance(optimizer, OptimizerV2)

        metadata["optimizer"].update(
            {
                str(step): {
                    "framework": "tensorflow",
                    "config": str(optimizer.get_config()),
                    "param_groups": [],
                }
            }
        )

    def _set_scheduler(self, metadata, scheduler, step):
        raise NotImplementedError()

    def _save_optimizer(self, step, optimizer):
        assert isinstance(optimizer, OptimizerV2)
        assert isinstance(self.config.metadata, dict)
        assert str(step) in self.config.metadata["optimizer"]

        for i in range(1, len(optimizer.weights)):
            param_name = optimizer.weights[i].name
            param_value = optimizer.weights[i].numpy()

            self._save_param(step, group="", name=param_name, weight=param_value)

            self.config.metadata["optimizer"][str(step)]["param_groups"].append(
                param_name
            )

    def _load_optimizer(self, step, layers, optimizer):
        assert isinstance(optimizer, OptimizerV2)
        assert isinstance(self.config.metadata, dict)

        step = str(step)
        if step not in self.config.metadata["optimizer"]:
            raise KeyError(
                "Available only in {}".format(
                    list(self.config.metadata["optimizer"].keys())
                )
            )

        _ordered_weight = {}
        for layer in layers:
            name = layer.object_name
            layer_image = self._client.get_object(
                bucket_name=self.config.bucket_name, object_name=name
            ).read()

            layer_image = h5py.File(io.BytesIO(layer_image), "r")

            name = name[name.find("/") + 1 :]
            _ordered_weight[name] = layer_image[self.type][:]

        new_weight = [np.array(int(step))]
        for param_name in self.config.metadata["optimizer"][str(step)]["param_groups"]:
            new_weight.append(_ordered_weight[param_name])

        _metadata_config = ast.literal_eval(
            self.config.metadata["optimizer"][str(step)]["config"]
        )
        optimizer.from_config(_metadata_config)
        optimizer.set_weights(new_weight)

    def save(self, optimizer):
        """
        save weight of optimizer

        Args:
            optimizer (:obj:`tf.keras.optimizers.Optimizer`, **require**):
                Tensorflow optimizer.

        Examples::

            >>> model = Model()
            # model training...
            >>> optimizer_manager.save(model.optimizer)

        """
        super(OptimizerManager, self).save(optimizer)

    def load(self, optimizer, step):
        """
        load weight of optimizer

        Args:
            optimizer (:obj:`tf.keras.optimizers.Optimizer`, **require**):
                Tensorflow optimizer.
            step (:obj:`integer`, **require**):
                optimizer step.

        Examples::

            >>> model = Model()
            >>> optimizer_manager.load(model.optimizer, step=938)

        """
        super(OptimizerManager, self).load(optimizer, step)
